from typing import Dict, Iterable, Optional, Union
import numpy as np
from .display import plot_hrir_plane, plot_hrtf_plane, plot_plane_angles, plot_hrir_lines, plot_hrtf_lines
from .full import Cipic, Ari, Listen, BiLi, CrossMod, Ita, Hutubs, Riec, Chedar, Widespread, Sadie2, Princeton3D3A, Scut, Sonicom, MitKemar
from .specifications import HrirPlaneSpec


class PlaneDatasetMixin:
    def __init__(self,
        plane: str,
        domain: str,
        side: str,
        plane_angles: Optional[Iterable[float]],
        plane_offset: float,
        positive_angles: bool,
        distance: Union[float, str],
        hrir_method: Optional[str],
        hrir_variant: str,
        hrir_offset: Optional[float],
        hrir_scaling: Optional[float],
        hrir_samplerate: Optional[float],
        hrir_length: Optional[float],
        hrir_min_phase: bool = False,
        hrir_role: str = 'features',
        other_specs: Optional[Dict] = None,
        **kwargs,
    ):
        hrirs_spec = HrirPlaneSpec(plane, domain, side, plane_angles, plane_offset, positive_angles, distance, hrir_method, hrir_variant, hrir_offset, hrir_scaling, hrir_samplerate, hrir_length, hrir_min_phase)
        if other_specs is None:
            other_specs = {}
        for spec_name in other_specs.keys():
            if hrir_role == spec_name.split('_spec')[0]:
                raise ValueError(f'No {spec_name} should be given since that role is already taken by the HRIRs')
        specs = {hrir_role+'_spec': hrirs_spec, **other_specs}
        super().__init__(**specs, **kwargs)
        if plane in ('horizontal', 'interaural'):
            self.plane_angle_name = 'yaw [°]'
        elif plane in ('median', 'vertical'):
            self.plane_angle_name = 'pitch [°]'
        else: # frontal plane
            self.plane_angle_name = 'roll [°]'


    @property
    def plane_angles(self):
        return self._plane_transform.plane_angles
    

    @property
    def positive_angles(self):
        return self._plane_transform.positive_angles


    @positive_angles.setter
    def positive_angles(self, value):
        self._plane_transform.positive_angles = value


    @property
    def min_angle(self):
        return self._plane_transform.min_angle


    @property
    def max_angle(self):
        return self._plane_transform.max_angle


    def plot_plane(self, idx, ax=None, vmin=None, vmax=None, title=None, lineplot=False, cmap='viridis', continuous=False, colorbar=True, log_freq=False):
        hrir_role = 'features' if 'hrir' in self._features_keys else 'target' if 'hrir' in self._target_keys else 'group'
        if vmin is None or vmax is None:
            all_hrirs = self[:][hrir_role]
            if vmin is None:
                vmin = all_hrirs.min()
            if vmax is None:
                vmax = all_hrirs.max()
        data = self[idx][hrir_role]

        if self._specification['hrir']['domain'] == 'time':
            if lineplot:
                ax = plot_hrir_lines(data, self.plane_angles, self.plane_angle_name, self.hrir_samplerate, ax=ax, vmin=vmin, vmax=vmax)
            else:
                ax = plot_hrir_plane(data, self.plane_angles, self.plane_angle_name, self.hrir_samplerate, ax=ax, vmin=vmin, vmax=vmax, cmap=cmap, continuous=continuous, colorbar=colorbar)
        else:
            if lineplot:
                ax = plot_hrtf_lines(data, self.plane_angles, self.plane_angle_name, self.hrtf_frequencies, log_freq=log_freq, ax=ax, vmin=vmin, vmax=vmax)
            else:
                ax = plot_hrtf_plane(data, self.plane_angles, self.plane_angle_name, self.hrtf_frequencies, log_freq=log_freq, ax=ax, vmin=vmin, vmax=vmax, cmap=cmap, continuous=continuous, colorbar=colorbar)

        if title is None:
            plane = self._specification['hrir']['plane']
            plane_offset = self._specification['hrir']['plane_offset']
            title = "{} Plane{} of Subject {}'s {} Ear".format(plane.title(),
                ' With Offset {}°'.format(plane_offset) if plane_offset != 0 or plane in ('vertical', 'interaural') else '',
                self.subject_ids[idx],
                self.sides[idx].replace('-', ' ').title(),
            )
        ax.set_title(title)
        return ax


    def plot_angles(self, ax=None, title=None):
        plane = self._specification['hrir']['plane']
        if plane in ('horizontal', 'interaural', 'frontal'):
            zero_location = 'N'
            direction = 'counterclockwise'
        else: # median or vertical
            zero_location = 'W'
            direction = 'clockwise'
        ax = plot_plane_angles(self.plane_angles, self.min_angle, self.max_angle, True, 1, zero_location, direction, ax) # TODO use actual radius
        if title is None:
            plane_offset = self._specification['hrir']['plane_offset']
            title = 'Angles in the {} Plane{}'.format(plane.title(),
                ' With Offset {}°'.format(plane_offset) if plane_offset != 0 or plane in ('vertical', 'interaural') else ''
            )
        ax.set_title(title)
        return ax


class CipicPlane(PlaneDatasetMixin, Cipic):
    def __init__(self,
        root: str,
        plane: str,
        domain: str = 'magnitude_db',
        side: str = 'left',
        plane_angles: Optional[Iterable[float]] = None,
        plane_offset: float = 0.,
        positive_angles: bool = False,
        hrir_offset: Optional[float] = None,
        hrir_scaling: Optional[float] = None,
        hrir_samplerate: Optional[float] = None,
        hrir_length: Optional[float] = None,
        hrir_min_phase: bool = False,
        hrir_role: str = 'features',
        other_specs: Optional[Dict] = None,
        subject_ids: Optional[Iterable[int]] = None,
        exclude_ids: Optional[Iterable[int]] = None,
        dtype: type = np.float32,
        download: bool = False,
        verify: bool = True,
    ):
        super().__init__(
            plane, domain, side, plane_angles, plane_offset, positive_angles, 'farthest', None, 'compensated', root=root,
            hrir_offset=hrir_offset, hrir_scaling=hrir_scaling, hrir_samplerate=hrir_samplerate, hrir_length=hrir_length, hrir_min_phase=hrir_min_phase,
            hrir_role=hrir_role, other_specs=other_specs, subject_ids=subject_ids, exclude_ids=exclude_ids, dtype=dtype,
            download=download, verify=verify,
        )


class AriPlane(PlaneDatasetMixin, Ari):
    def __init__(self,
        root: str,
        plane: str,
        domain: str = 'magnitude_db',
        side: str = 'left',
        plane_angles: Optional[Iterable[float]] = None,
        plane_offset: float = 0.,
        positive_angles: bool = False,
        hrir_offset: Optional[float] = None,
        hrir_scaling: Optional[float] = None,
        hrir_samplerate: Optional[float] = None,
        hrir_length: Optional[float] = None,
        hrir_min_phase: bool = False,
        hrir_role: str = 'features',
        other_specs: Optional[Dict] = None,
        subject_ids: Optional[Iterable[int]] = None,
        exclude_ids: Optional[Iterable[int]] = None,
        dtype: type = np.float32,
        download: bool = False,
        verify: bool = True,
    ):
        if positive_angles is None:
            positive_angles = False
        super().__init__(
            plane, domain, side, plane_angles, plane_offset, positive_angles, 'farthest', None, 'compensated', root=root,
            hrir_offset=hrir_offset, hrir_scaling=hrir_scaling, hrir_samplerate=hrir_samplerate, hrir_length=hrir_length, hrir_min_phase=hrir_min_phase,
            hrir_role=hrir_role, other_specs=other_specs, subject_ids=subject_ids, exclude_ids=exclude_ids, dtype=dtype,
            download=download, verify=verify,
        )


class ListenPlane(PlaneDatasetMixin, Listen):
    def __init__(self,
        root: str,
        plane: str,
        domain: str = 'magnitude_db',
        side: str = 'left',
        plane_angles: Optional[Iterable[float]] = None,
        plane_offset: float = 0.,
        positive_angles: bool = False,
        hrir_variant: str = 'compensated',
        hrir_offset: Optional[float] = None,
        hrir_scaling: Optional[float] = None,
        hrir_samplerate: Optional[float] = None,
        hrir_length: Optional[float] = None,
        hrir_min_phase: bool = False,
        hrir_role: str = 'features',
        other_specs: Optional[Dict] = None,
        subject_ids: Optional[Iterable[int]] = None,
        exclude_ids: Optional[Iterable[int]] = None,
        dtype: type = np.float32,
        download: bool = False,
        verify: bool = True,
    ):
        super().__init__(
            plane, domain, side, plane_angles, plane_offset, positive_angles, 'farthest', None, hrir_variant, root=root,
            hrir_offset=hrir_offset, hrir_scaling=hrir_scaling, hrir_samplerate=hrir_samplerate, hrir_length=hrir_length, hrir_min_phase=hrir_min_phase,
            hrir_role=hrir_role, other_specs=other_specs, subject_ids=subject_ids, exclude_ids=exclude_ids, dtype=dtype,
            download=download, verify=verify,
        )


class CrossModPlane(PlaneDatasetMixin, CrossMod):
    def __init__(self,
        root: str,
        plane: str,
        domain: str = 'magnitude_db',
        side: str = 'left',
        plane_angles: Optional[Iterable[float]] = None,
        plane_offset: float = 0.,
        positive_angles: bool = False,
        hrir_variant: str = 'compensated',
        hrir_offset: Optional[float] = None,
        hrir_scaling: Optional[float] = None,
        hrir_samplerate: Optional[float] = None,
        hrir_length: Optional[float] = None,
        hrir_min_phase: bool = False,
        hrir_role: str = 'features',
        other_specs: Optional[Dict] = None,
        subject_ids: Optional[Iterable[int]] = None,
        exclude_ids: Optional[Iterable[int]] = None,
        dtype: type = np.float32,
        download: bool = False,
        verify: bool = True,
    ):
        super().__init__(
            plane, domain, side, plane_angles, plane_offset, positive_angles, 'farthest', None, hrir_variant, root=root,
            hrir_offset=hrir_offset, hrir_scaling=hrir_scaling, hrir_samplerate=hrir_samplerate, hrir_length=hrir_length, hrir_min_phase=hrir_min_phase,
            hrir_role=hrir_role, other_specs=other_specs, subject_ids=subject_ids, exclude_ids=exclude_ids, dtype=dtype,
            download=download, verify=verify,
        )


class BiLiPlane(PlaneDatasetMixin, BiLi):
    def __init__(self,
        root: str,
        plane: str,
        domain: str = 'magnitude_db',
        side: str = 'left',
        plane_angles: Optional[Iterable[float]] = None,
        plane_offset: float = 0.,
        positive_angles: bool = False,
        hrir_variant: str = 'compensated',
        hrir_offset: Optional[float] = None,
        hrir_scaling: Optional[float] = None,
        hrir_samplerate: Optional[float] = None,
        hrir_length: Optional[float] = None,
        hrir_min_phase: bool = False,
        hrir_role: str = 'features',
        other_specs: Optional[Dict] = None,
        subject_ids: Optional[Iterable[int]] = None,
        exclude_ids: Optional[Iterable[int]] = None,
        dtype: type = np.float32,
        download: bool = False,
        verify: bool = True,
    ):
        super().__init__(
            plane, domain, side, plane_angles, plane_offset, positive_angles, 'farthest', None, hrir_variant, root=root,
            hrir_offset=hrir_offset, hrir_scaling=hrir_scaling, hrir_samplerate=hrir_samplerate, hrir_length=hrir_length, hrir_min_phase=hrir_min_phase,
            hrir_role=hrir_role, other_specs=other_specs, subject_ids=subject_ids, exclude_ids=exclude_ids, dtype=dtype,
            download=download, verify=verify,
        )


class ItaPlane(PlaneDatasetMixin, Ita):
    def __init__(self,
        root: str,
        plane: str,
        domain: str = 'magnitude_db',
        side: str = 'left',
        plane_angles: Optional[Iterable[float]] = None,
        plane_offset: float = 0.,
        positive_angles: bool = False,
        hrir_offset: Optional[float] = None,
        hrir_scaling: Optional[float] = None,
        hrir_samplerate: Optional[float] = None,
        hrir_length: Optional[float] = None,
        hrir_min_phase: bool = False,
        hrir_role: str = 'features',
        other_specs: Optional[Dict] = None,
        subject_ids: Optional[Iterable[int]] = None,
        exclude_ids: Optional[Iterable[int]] = None,
        dtype: type = np.float32,
        download: bool = False,
        verify: bool = True,
    ):
        super().__init__(
            plane, domain, side, plane_angles, plane_offset, positive_angles, 'farthest', None, 'compensated', root=root,
            hrir_offset=hrir_offset, hrir_scaling=hrir_scaling, hrir_samplerate=hrir_samplerate, hrir_length=hrir_length, hrir_min_phase=hrir_min_phase,
            hrir_role=hrir_role, other_specs=other_specs, subject_ids=subject_ids, exclude_ids=exclude_ids, dtype=dtype,
            download=download, verify=verify,
        )


class HutubsPlane(PlaneDatasetMixin, Hutubs):
    def __init__(self,
        root: str,
        plane: str,
        domain: str = 'magnitude_db',
        side: str = 'left',
        plane_angles: Optional[Iterable[float]] = None,
        plane_offset: float = 0.,
        positive_angles: bool = False,
        hrir_method: Optional[str] = 'acoustic',
        hrir_variant: str = 'compensated',
        hrir_offset: Optional[float] = None,
        hrir_scaling: Optional[float] = None,
        hrir_samplerate: Optional[float] = None,
        hrir_length: Optional[float] = None,
        hrir_min_phase: bool = False,
        hrir_role: str = 'features',
        other_specs: Optional[Dict] = None,
        subject_ids: Optional[Iterable[int]] = None,
        exclude_ids: Optional[Iterable[int]] = None,
        dtype: type = np.float32,
        download: bool = False,
        verify: bool = True,
    ):
        super().__init__(
            plane, domain, side, plane_angles, plane_offset, positive_angles, 'farthest', hrir_method, hrir_variant, root=root,
            hrir_offset=hrir_offset, hrir_scaling=hrir_scaling, hrir_samplerate=hrir_samplerate, hrir_length=hrir_length, hrir_min_phase=hrir_min_phase,
            hrir_role=hrir_role, other_specs=other_specs, subject_ids=subject_ids, exclude_ids=exclude_ids, dtype=dtype,
            download=download, verify=verify,
        )


class RiecPlane(PlaneDatasetMixin, Riec):
    def __init__(self,
        root: str,
        plane: str,
        domain: str = 'magnitude_db',
        side: str = 'left',
        plane_angles: Optional[Iterable[float]] = None,
        plane_offset: float = 0.,
        positive_angles: bool = False,
        distance: Union[float, str] = 'farthest',
        hrir_offset: Optional[float] = None,
        hrir_scaling: Optional[float] = None,
        hrir_samplerate: Optional[float] = None,
        hrir_length: Optional[float] = None,
        hrir_min_phase: bool = False,
        hrir_role: str = 'features',
        other_specs: Optional[Dict] = None,
        subject_ids: Optional[Iterable[int]] = None,
        exclude_ids: Optional[Iterable[int]] = None,
        dtype: type = np.float32,
        download: bool = False,
        verify: bool = True,
    ):
        super().__init__(
            plane, domain, side, plane_angles, plane_offset, positive_angles, distance, None, 'compensated', root=root,
            hrir_offset=hrir_offset, hrir_scaling=hrir_scaling, hrir_samplerate=hrir_samplerate, hrir_length=hrir_length, hrir_min_phase=hrir_min_phase,
            hrir_role=hrir_role, other_specs=other_specs, subject_ids=subject_ids, exclude_ids=exclude_ids, dtype=dtype,
            download=download, verify=verify,
        )


class ChedarPlane(PlaneDatasetMixin, Chedar):
    def __init__(self,
        root: str,
        plane: str,
        domain: str = 'magnitude_db',
        side: str = 'left',
        plane_angles: Optional[Iterable[float]] = None,
        plane_offset: float = 0.,
        positive_angles: bool = False,
        distance: Union[float, str] = 'farthest',
        hrir_offset: Optional[float] = None,
        hrir_scaling: Optional[float] = None,
        hrir_samplerate: Optional[float] = None,
        hrir_length: Optional[float] = None,
        hrir_min_phase: bool = False,
        hrir_role: str = 'features',
        other_specs: Optional[Dict] = None,
        subject_ids: Optional[Iterable[int]] = None,
        exclude_ids: Optional[Iterable[int]] = None,
        dtype: type = np.float32,
        download: bool = False,
        verify: bool = False,
    ):
        super().__init__(
            plane, domain, side, plane_angles, plane_offset, positive_angles, distance, None, 'compensated', root=root,
            hrir_offset=hrir_offset, hrir_scaling=hrir_scaling, hrir_samplerate=hrir_samplerate, hrir_length=hrir_length, hrir_min_phase=hrir_min_phase,
            hrir_role=hrir_role, other_specs=other_specs, subject_ids=subject_ids, exclude_ids=exclude_ids, dtype=dtype,
            download=download, verify=verify,
        )


class WidespreadPlane(PlaneDatasetMixin, Widespread):
    def __init__(self,
        root: str,
        plane: str,
        domain: str = 'magnitude_db',
        side: str = 'left',
        plane_angles: Optional[Iterable[float]] = None,
        plane_offset: float = 0.,
        positive_angles: bool = False,
        distance: Union[float, str] = 'farthest',
        hrir_offset: Optional[float] = None,
        hrir_scaling: Optional[float] = None,
        hrir_samplerate: Optional[float] = None,
        hrir_length: Optional[float] = None,
        hrir_min_phase: bool = False,
        hrir_role: str = 'features',
        other_specs: Optional[Dict] = None,
        subject_ids: Optional[Iterable[int]] = None,
        exclude_ids: Optional[Iterable[int]] = None,
        grid: str = 'UV',
        dtype: type = np.float32,
        download: bool = False,
        verify: bool = False,
    ):
        super().__init__(
            plane, domain, side, plane_angles, plane_offset, positive_angles, distance, None, 'compensated', root=root,
            hrir_offset=hrir_offset, hrir_scaling=hrir_scaling, hrir_samplerate=hrir_samplerate, hrir_length=hrir_length, hrir_min_phase=hrir_min_phase,
            hrir_role=hrir_role, other_specs=other_specs, subject_ids=subject_ids, exclude_ids=exclude_ids, dtype=dtype, grid=grid,
            download=download, verify=verify,
        )


class Sadie2Plane(PlaneDatasetMixin, Sadie2):
    def __init__(self,
        root: str,
        plane: str,
        domain: str = 'magnitude_db',
        side: str = 'left',
        plane_angles: Optional[Iterable[float]] = None,
        plane_offset: float = 0.,
        positive_angles: bool = False,
        hrir_offset: Optional[float] = None,
        hrir_scaling: Optional[float] = None,
        hrir_samplerate: Optional[float] = None,
        hrir_length: Optional[float] = None,
        hrir_min_phase: bool = False,
        hrir_role: str = 'features',
        other_specs: Optional[Dict] = None,
        subject_ids: Optional[Iterable[int]] = None,
        exclude_ids: Optional[Iterable[int]] = None,
        dtype: type = np.float32,
        download: bool = False,
        verify: bool = True,
    ):
        super().__init__(
            plane, domain, side, plane_angles, plane_offset, positive_angles, 'farthest', None, 'compensated', root=root,
            hrir_offset=hrir_offset, hrir_scaling=hrir_scaling, hrir_samplerate=hrir_samplerate, hrir_length=hrir_length, hrir_min_phase=hrir_min_phase,
            hrir_role=hrir_role, other_specs=other_specs, subject_ids=subject_ids, exclude_ids=exclude_ids, dtype=dtype,
            download=download, verify=verify,
        )


class Princeton3D3APlane(PlaneDatasetMixin, Princeton3D3A):
    def __init__(self,
        root: str,
        plane: str,
        domain: str = 'magnitude_db',
        side: str = 'left',
        plane_angles: Optional[Iterable[float]] = None,
        plane_offset: float = 0.,
        positive_angles: bool = False,
        hrir_method: Optional[str] = 'acoustic',
        hrir_variant: str = 'compensated',
        hrir_offset: Optional[float] = None,
        hrir_scaling: Optional[float] = None,
        hrir_samplerate: Optional[float] = None,
        hrir_length: Optional[float] = None,
        hrir_min_phase: bool = False,
        hrir_role: str = 'features',
        other_specs: Optional[Dict] = None,
        subject_ids: Optional[Iterable[int]] = None,
        exclude_ids: Optional[Iterable[int]] = None,
        dtype: type = np.float32,
        download: bool = False,
        verify: bool = True,
    ):
        super().__init__(
            plane, domain, side, plane_angles, plane_offset, positive_angles, 'farthest', hrir_method, hrir_variant, root=root,
            hrir_offset=hrir_offset, hrir_scaling=hrir_scaling, hrir_samplerate=hrir_samplerate, hrir_length=hrir_length, hrir_min_phase=hrir_min_phase,
            hrir_role=hrir_role, other_specs=other_specs, subject_ids=subject_ids, exclude_ids=exclude_ids, dtype=dtype,
            download=download, verify=verify,
        )


class ScutPlane(PlaneDatasetMixin, Scut):
    def __init__(self,
        root: str,
        plane: str,
        domain: str = 'magnitude_db',
        side: str = 'left',
        plane_angles: Optional[Iterable[float]] = None,
        plane_offset: float = 0.,
        positive_angles: bool = False,
        distance: Union[float, str] = 'farthest',
        hrir_method: Optional[str] = 'acoustic',
        hrir_offset: Optional[float] = None,
        hrir_scaling: Optional[float] = None,
        hrir_samplerate: Optional[float] = None,
        hrir_length: Optional[float] = None,
        hrir_min_phase: bool = False,
        hrir_role: str = 'features',
        other_specs: Optional[Dict] = None,
        subject_ids: Optional[Iterable[int]] = None,
        exclude_ids: Optional[Iterable[int]] = None,
        dtype: type = np.float32,
        download: bool = False,
        verify: bool = True,
    ):
        super().__init__(
            plane, domain, side, plane_angles, plane_offset, positive_angles, distance, hrir_method, 'compensated', root=root,
            hrir_offset=hrir_offset, hrir_scaling=hrir_scaling, hrir_samplerate=hrir_samplerate, hrir_length=hrir_length, hrir_min_phase=hrir_min_phase,
            hrir_role=hrir_role, other_specs=other_specs, subject_ids=subject_ids, exclude_ids=exclude_ids, dtype=dtype,
            download=download, verify=verify,
        )


class SonicomPlane(PlaneDatasetMixin, Sonicom):
    def __init__(self,
        root: str,
        plane: str,
        domain: str = 'magnitude_db',
        side: str = 'left',
        plane_angles: Optional[Iterable[float]] = None,
        plane_offset: float = 0.,
        positive_angles: bool = False,
        hrir_variant: str = 'compensated',
        hrir_offset: Optional[float] = None,
        hrir_scaling: Optional[float] = None,
        hrir_samplerate: Optional[float] = None,
        hrir_length: Optional[float] = None,
        hrir_min_phase: bool = False,
        hrir_role: str = 'features',
        other_specs: Optional[Dict] = None,
        subject_ids: Optional[Iterable[int]] = None,
        exclude_ids: Optional[Iterable[int]] = None,
        dtype: type = np.float32,
        download: bool = False,
        verify: bool = True,
    ):
        super().__init__(
            plane, domain, side, plane_angles, plane_offset, positive_angles, 'farthest', None, hrir_variant, root=root,
            hrir_offset=hrir_offset, hrir_scaling=hrir_scaling, hrir_samplerate=hrir_samplerate, hrir_length=hrir_length, hrir_min_phase=hrir_min_phase,
            hrir_role=hrir_role, other_specs=other_specs, subject_ids=subject_ids, exclude_ids=exclude_ids, dtype=dtype,
            download=download, verify=verify,
        )


class MitKemarPlane(PlaneDatasetMixin, MitKemar):
    def __init__(self,
        root: str,
        plane: str,
        domain: str = 'magnitude_db',
        side: str = 'left',
        plane_angles: Optional[Iterable[float]] = None,
        plane_offset: float = 0.,
        positive_angles: bool = False,
        distance: Union[float, str] = 'farthest',
        hrir_offset: Optional[float] = None,
        hrir_scaling: Optional[float] = None,
        hrir_samplerate: Optional[float] = None,
        hrir_length: Optional[float] = None,
        hrir_min_phase: bool = False,
        hrir_role: str = 'features',
        other_specs: Optional[Dict] = None,
        subject_ids: Optional[Iterable[int]] = None,
        exclude_ids: Optional[Iterable[int]] = None,
        dtype: type = np.float32,
        download: bool = False,
        verify: bool = True,
    ):
        super().__init__(
            plane, domain, side, plane_angles, plane_offset, positive_angles, distance, None, 'compensated', root=root,
            hrir_offset=hrir_offset, hrir_scaling=hrir_scaling, hrir_samplerate=hrir_samplerate, hrir_length=hrir_length, hrir_min_phase=hrir_min_phase,
            hrir_role=hrir_role, other_specs=other_specs, subject_ids=subject_ids, exclude_ids=exclude_ids, dtype=dtype,
            download=download, verify=verify,
        )



"""A Python toolkit for data-driven HRTF research"""

__version__ = '0.6.3'

from .full import Dataset, Cipic, Ari, Listen, CrossMod, BiLi, Ita, Hutubs, Riec, Sadie2, Princeton3D3A, Chedar, Widespread, Scut, Sonicom, MitKemar
from .planar import CipicPlane, AriPlane, ListenPlane, CrossModPlane, BiLiPlane, ItaPlane, HutubsPlane, RiecPlane, Sadie2Plane, Princeton3D3APlane, ChedarPlane, WidespreadPlane, ScutPlane, SonicomPlane, MitKemarPlane
from .specifications import CollectionSpec, SideSpec, SubjectSpec, HrirSpec, HrirPlaneSpec, AnthropometrySpec, ImageSpec, MeshSpec
from .transforms import BatchTransform



import numpy as np
from typing import Iterable


def wrap_closed_open_interval(values, lower, upper):
    return (np.asarray(values) - lower) % (upper - lower) + lower


def wrap_open_closed_interval(values, lower, upper):
    return -((lower - np.asarray(values)) % (upper - lower)) - lower


def wrap_closed_interval(values, lower, upper):
    values = np.asarray(values)
    return np.where((values < lower) | (values > upper), wrap_closed_open_interval(values, lower, upper), values)


def spherical2interaural(azimuth, elevation, radius, angles_in_degrees=True):
    if angles_in_degrees:
        azimuth = np.deg2rad(azimuth)
        elevation = np.deg2rad(elevation)
    lateral = np.arcsin(np.sin(azimuth) * np.cos(elevation))
    vertical = np.arctan2(np.tan(elevation), np.cos(azimuth))
    if angles_in_degrees:
        return np.rad2deg(lateral), np.rad2deg(vertical), radius
    else:
        return lateral, vertical, radius


def spherical2cartesian(azimuth, elevation, radius, angles_in_degrees=True):
    if angles_in_degrees:
        azimuth = np.deg2rad(azimuth)
        elevation = np.deg2rad(elevation)
    hoz_proj = radius * np.cos(elevation)
    x = hoz_proj * np.cos(azimuth)
    y = hoz_proj * np.sin(azimuth)
    z = radius * np.sin(elevation)
    return x, y, z


def cartesian2spherical(x, y, z, angles_in_degrees=True):
    x = np.asarray(x)
    y = np.asarray(y)
    z = np.asarray(z)
    azimuth = np.arctan2(y, x)
    elevation = np.arctan2(z, np.sqrt(x**2 + y**2))
    radius = np.sqrt(x**2 + y**2 + z**2)
    if angles_in_degrees:
        return np.rad2deg(azimuth), np.rad2deg(elevation), radius
    else:
        return azimuth, elevation, radius


def cartesian2interaural(x, y, z, angles_in_degrees=True):
    x = np.asarray(x)
    y = np.asarray(y)
    z = np.asarray(z)
    lateral = np.arctan2(y, np.sqrt(x**2 + z**2))
    vertical = np.arctan2(z, x) 
    radius = np.sqrt(x**2 + y**2 + z**2)
    if angles_in_degrees:
        return np.rad2deg(lateral), np.rad2deg(vertical), radius
    else:
        return lateral, vertical, radius


def interaural2spherical(lateral, vertical, radius, angles_in_degrees=True):
    if angles_in_degrees:
        lateral = np.deg2rad(lateral)
        vertical = np.deg2rad(vertical)
    azimuth = np.arctan2(np.tan(lateral), np.cos(vertical))
    elevation = np.arcsin(np.cos(lateral) * np.sin(vertical))
    if angles_in_degrees:
        return np.rad2deg(azimuth), np.rad2deg(elevation), radius
    else:
        return azimuth, elevation, radius


def interaural2cartesian(lateral, vertical, radius, angles_in_degrees=True):
    if angles_in_degrees:
        lateral = np.deg2rad(lateral)
        vertical = np.deg2rad(vertical)
    med_proj = radius * np.cos(lateral)
    x = med_proj * np.cos(vertical)
    y = radius * np.sin(lateral)
    z = med_proj * np.sin(vertical)
    return x, y, z


def azimuth_elevation_from_yaw(yaw_angles, plane_offset=0):
    norm_yaw = wrap_closed_open_interval(yaw_angles, -180, 180)
    try:
        return tuple(norm_yaw), (plane_offset,) * len(norm_yaw)
    except TypeError:
        return (norm_yaw,), (plane_offset,)


def azimuth_elevation_from_pitch(pitch_angles, plane_offset=0):
    norm_pitch = wrap_closed_open_interval(pitch_angles, -90, 270)
    azimuth_angles = np.where(norm_pitch < 90, plane_offset, plane_offset - 180)
    elevation_angles = np.where(norm_pitch < 90, norm_pitch, 180 - norm_pitch)
    try:
        return tuple(azimuth_angles), tuple(elevation_angles)
    except TypeError:
        return (azimuth_angles.item(),), (elevation_angles.item(),)


def azimuth_elevation_from_roll(roll_angles, plane_offset=0):
    norm_roll = wrap_closed_open_interval(roll_angles, -180, 180)
    azimuth_angles = np.where(norm_roll < 0, plane_offset + 90, plane_offset - 90)
    elevation_angles = np.where(norm_roll < 0, norm_roll + 90, 90 - norm_roll)
    try:
        return tuple(azimuth_angles), tuple(elevation_angles)
    except TypeError:
        return (azimuth_angles.item(),), (elevation_angles.item(),)


def lateral_vertical_from_yaw(yaw_angles, plane_offset=0):
    norm_yaw = wrap_closed_open_interval(yaw_angles, -90, 270)
    lateral_angles = np.where(norm_yaw < 90, norm_yaw, 180 - norm_yaw)
    vertical_angles = np.where(norm_yaw < 90, plane_offset, plane_offset - 180)
    try:
        return tuple(lateral_angles), tuple(vertical_angles)
    except TypeError:
        return (lateral_angles.item(),), (vertical_angles.item(),)


def lateral_vertical_from_pitch(pitch_angles, plane_offset=0):
    norm_pitch = wrap_closed_open_interval(pitch_angles, -180, 180)
    try:
        return (plane_offset,) * len(norm_pitch), tuple(norm_pitch)
    except TypeError:
        return (plane_offset,), (norm_pitch,)


def lateral_vertical_from_roll(roll_angles, plane_offset=0):
    norm_roll = wrap_closed_open_interval(roll_angles, -90, 270)
    lateral_angles = np.where(norm_roll < 90, -norm_roll, norm_roll - 180)
    vertical_angles = np.where(norm_roll < 90, plane_offset + 90, plane_offset - 90)
    try:
        return tuple(lateral_angles), tuple(vertical_angles)
    except TypeError:
        return (lateral_angles.item(),), (vertical_angles.item(),)


def quantise(values, precision):
    if precision > 0:
        return (values / precision).round(decimals=0) * precision
    else:
        return values.round(decimals=-precision)



from typing import Optional
from PIL.Image import Image, LANCZOS


class CenterCropResizeGreyscale:
    def __init__(self, resize_width: Optional[int], resize_height: Optional[int], crop_width: Optional[int]=None, crop_height: Optional[int]=None):
        self.resize_width = resize_width
        self.resize_height = resize_height
        self.crop_width = crop_width
        self.crop_height = crop_height


    def __call__(self, image: Image) -> Image:
        orig_width, orig_height = image.size
        if self.resize_width is None:
            resize_width = orig_width if self.crop_width is None else self.crop_width
        else:
            resize_width = self.resize_width
        if self.resize_height is None:
            resize_height = orig_height if self.crop_height is None else self.crop_height
        else:
            resize_height = self.resize_height
        if self.crop_width is not None or self.crop_height is not None:
            if self.crop_width is not None:
                box_left = (orig_width - self.crop_width) // 2
                box_right = box_left + self.crop_width
            else:
                box_left = 0
                box_right = orig_width
            if self.crop_height is not None:
                box_top = (orig_height - self.crop_height) // 2
                box_bottom = box_top + self.crop_height
            else:
                box_top = 0
                box_bottom = orig_height
            box = (box_left, box_top, box_right, box_bottom)
        else:
            box = None
        return image.resize((resize_width, resize_height), resample=LANCZOS, box=box).convert('L')



from .query import DataQuery, CipicDataQuery, AriDataQuery, ListenDataQuery, BiLiDataQuery, CrossModDataQuery, ItaDataQuery, HutubsDataQuery, RiecDataQuery, ChedarDataQuery, WidespreadDataQuery, Sadie2DataQuery, Princeton3D3ADataQuery, ScutDataQuery, SonicomDataQuery, MitKemarDataQuery
from .util import wrap_closed_open_interval, wrap_closed_interval, spherical2cartesian, spherical2interaural, cartesian2spherical, cartesian2interaural, interaural2spherical, interaural2cartesian, quantise
from abc import abstractmethod
from pathlib import Path
from typing import Optional, Union
import numpy as np
import netCDF4 as ncdf
from PIL import Image


class DataReader:

    def __init__(self,
        query: DataQuery,
    ):
        self.query = query

    
    def collection(self, suffix=None):
        if suffix is None:
            return self.query.collection_id
        return f'{self.query.collection_id}-{suffix}'


class SofaDataReader(DataReader):
    """
    An abstract class that reads the HRIR for a given subject id of a dataset from a SOFA file and stores it internally
    as a 3D tensor. The HRIR directions are stored in the rows and columns as a plate carrée projection where each
    column represents a plane parallel to the fundamental plane, i.e. each row represents a single angle in the
    fundamental plane. The poles (if present) are therefore stored in the first and last column.

    fundamental_angle: angle in the fundamental plane with range  [-180, 180)
    (azimuth for spherical coordinates, vertical angle for interaural coordinates)
    orthogonal_angle: angle between fundamental plane and directional vector with range [-90, 90]
    (elevation for spherical coordinates, lateral angle for interaural coordinates)
    """

    _angle_quantisation: int = -3
    _distance_quantisation: int = -3


    @property
    @abstractmethod
    def fundamental_angle_name(self):
        pass


    @property
    @abstractmethod
    def orthogonal_angle_name(self):
        pass


    @abstractmethod
    def _sofa_path(self, subject_id):
        pass


    @staticmethod
    @abstractmethod
    def _convert_positions(coordinate_system, positions):
        pass


    @staticmethod
    @abstractmethod
    def _coordinate_transform(coordinate_system, selected_fundamental_angles, selected_orthogonal_angles, selected_radii):
        pass


    @staticmethod
    @abstractmethod
    def _mirror_hrirs(hrirs, selected_fundamental_angles):
        pass


    @staticmethod
    @abstractmethod
    def _verify_angle_symmetry(fundamental_angles, orthogonal_angles):
        pass


    def hrir_samplerate(self, subject_id):
        sofa_path = self._sofa_path(subject_id)
        hrir_file = ncdf.Dataset(sofa_path)
        try:
            samplerate = hrir_file.variables['Data.SamplingRate'][:].item()
        except Exception as exc:
            raise ValueError(f'Error reading file "{sofa_path}"') from exc
        finally:
            hrir_file.close()
        return samplerate


    def hrir_length(self, subject_id):
        sofa_path = self._sofa_path(subject_id)
        hrir_file = ncdf.Dataset(sofa_path)
        try:
            length = hrir_file.dimensions['N'].size
        except Exception as exc:
            raise ValueError(f'Error reading file "{sofa_path}"') from exc
        finally:
            hrir_file.close()
        return length


    def _map_sofa_position_order_to_matrix(self, subject_id, fundamental_angles, orthogonal_angles, distance):
        if fundamental_angles is not None and orthogonal_angles is not None and len(fundamental_angles) != len(orthogonal_angles):
            raise ValueError(f'The number of fundamental angles ({len(fundamental_angles)}) differs from the number of orthogonal angles ({len(orthogonal_angles)})')
        sofa_path = self._sofa_path(subject_id)
        hrir_file = ncdf.Dataset(sofa_path)
        try:
            positions = self._convert_positions(
                hrir_file.variables['SourcePosition'].Type,
                np.ma.getdata(hrir_file.variables['SourcePosition'][:]),
            )
        except Exception as exc:
            raise ValueError(f'Error reading file "{sofa_path}"') from exc
        finally:
            hrir_file.close()
        quantised_positions =  np.column_stack((quantise(positions[:, :2], self._angle_quantisation), 
                                                quantise(positions[:, 2], self._distance_quantisation)))
        quantised_positions[:, 0] = wrap_closed_open_interval(quantised_positions[:, 0], -180, 180)
        if isinstance(distance, str):
            file_radii = np.unique(quantised_positions[:, 2])
            if distance == 'nearest':
                distance = file_radii[0]
            elif distance == 'farthest':
                distance = file_radii[-1]
            else:
                raise ValueError(f'Invalid distance "{distance}". Only "nearest", "farthest" or a value in meter is allowed.')

        if fundamental_angles is not None:
            fundamental_angles = wrap_closed_open_interval(fundamental_angles, -180, 180)
        if orthogonal_angles is not None:
            orthogonal_angles = wrap_closed_interval(orthogonal_angles, -90, 90)
        if fundamental_angles is None and orthogonal_angles is None and distance is None:
            # Read all positions in file, without extra checks
            selected_file_indices = list(range(len(quantised_positions)))
        else:
            # Check if file positions are part of requested positions
            selected_file_indices = []
            if fundamental_angles is None:
                check_fundamental_angle = lambda _: True
            else:
                check_fundamental_angle = lambda file_fundamental_angle: np.isclose(file_fundamental_angle, fundamental_angles)
            if orthogonal_angles is None:
                check_orthogonal_angle = lambda _: True
            else:
                check_orthogonal_angle = lambda file_orthogonal_angle: np.isclose(file_orthogonal_angle, orthogonal_angles)
            if distance is None:
                check_distance = lambda _: True
            else:
                check_distance = lambda file_radius: np.isclose(file_radius, distance)
            for file_idx, (file_fundamental_angle, file_orthogonal_angle, file_radius) in enumerate(quantised_positions):
                if (check_fundamental_angle(file_fundamental_angle) & check_orthogonal_angle(file_orthogonal_angle) & check_distance(file_radius)).any():
                    selected_file_indices.append(file_idx)

        selected_positions = quantised_positions[selected_file_indices]
        selected_fundamental_angles = selected_positions[:, 0]
        selected_orthogonal_angles = selected_positions[:, 1]
        selected_radii = selected_positions[:, 2]
        unique_radii = np.unique(selected_radii)

        for pole_angle in (-90, 90):
            # If value at pole requested
            if orthogonal_angles is None or np.isclose(orthogonal_angles, pole_angle).any():
                for radius in unique_radii:
                    pole_indices = np.flatnonzero(np.isclose(quantised_positions[:, 1], pole_angle) & (quantised_positions[:, 2] == radius))
                    # If at least one value at pole present in file
                    if len(pole_indices) > 0:
                        # Make sure to include all requested row angles at pole
                        if fundamental_angles is not None and orthogonal_angles is not None:
                            requested_fundamental_angles_at_pole = fundamental_angles[orthogonal_angles == pole_angle]
                            selected_fundamental_angles = np.concatenate((selected_fundamental_angles, requested_fundamental_angles_at_pole))
                        # If pole angle not present in selection yet
                        if len(selected_orthogonal_angles) == 0 or not np.isclose(selected_orthogonal_angles, pole_angle).any():
                            # Add to column angles at appropriate extremum
                            if pole_angle > 0 or len(selected_orthogonal_angles) == 0:
                                selected_orthogonal_angles = np.append(selected_orthogonal_angles, pole_angle)
                            else:
                                selected_orthogonal_angles = np.insert(selected_orthogonal_angles, 0, pole_angle)

        unique_fundamental_angles = np.unique(selected_fundamental_angles)
        unique_orthogonal_angles = np.unique(selected_orthogonal_angles)

        additional_pole_indices = []

        def repeat_value_at_pole(pole_angle, pole_column_idx):
            # If value at pole requested
            if orthogonal_angles is None or np.isclose(orthogonal_angles, pole_angle).any():
                for radius_idx, radius in enumerate(unique_radii):
                    pole_indices = np.flatnonzero(np.isclose(quantised_positions[:, 1], pole_angle) & (quantised_positions[:, 2] == radius))
                    # If at least one value at pole present in file
                    if len(pole_indices) > 0:
                        pole_idx = pole_indices[0]
                        # Copy to those row angles that miss the value at the pole (if any)
                        radius_positions = selected_positions[selected_positions[:, 2] == radius]
                        present_fundamental_angles = radius_positions[np.isclose(radius_positions[:, 1], pole_angle), 0]
                        missing_fundamental_angles = np.setdiff1d(unique_fundamental_angles, present_fundamental_angles)
                        missing_row_indices = [np.argmax(angle == unique_fundamental_angles) for angle in missing_fundamental_angles]
                        selected_file_indices.extend([pole_idx] * len(missing_row_indices))
                        additional_pole_indices.extend([(row_idx, pole_column_idx, radius_idx) for row_idx in missing_row_indices])

        repeat_value_at_pole(-90, 0)
        repeat_value_at_pole(90, -1)

        selection_mask_indices = []
        for file_fundamental_angle, file_orthogonal_angle, file_radius in selected_positions:
            selection_mask_indices.append((np.argmax(file_fundamental_angle == unique_fundamental_angles), np.argmax(file_orthogonal_angle == unique_orthogonal_angles), np.argmax(file_radius == unique_radii)))
        selection_mask_indices.extend(additional_pole_indices)
        if len(selection_mask_indices) == 0:
            raise ValueError('None of the specified HRIR measurement positions are available in this dataset')

        selection_mask_indices = tuple(np.array(selection_mask_indices).T)
        selection_mask = np.full((len(unique_fundamental_angles), len(unique_orthogonal_angles), len(unique_radii)), True)
        selection_mask[selection_mask_indices] = False

        return unique_fundamental_angles, unique_orthogonal_angles, unique_radii, selection_mask, selected_file_indices, selection_mask_indices


    def hrir_positions(self, subject_id, coordinate_system, fundamental_angles=None, orthogonal_angles=None, distance=None):
        selected_fundamental_angles, selected_orthogonal_angles, selected_radii, selection_mask, *_ = self._map_sofa_position_order_to_matrix(subject_id, fundamental_angles, orthogonal_angles, distance)

        coordinates = self._coordinate_transform(coordinate_system, selected_fundamental_angles, selected_orthogonal_angles, selected_radii)

        position_grid = np.stack(np.meshgrid(*coordinates, indexing='ij'), axis=-1)
        if selection_mask.any(): # sparse grid
            tiled_position_mask = np.tile(selection_mask[:, :, :, np.newaxis], (1, 1, 1, 3))
            return np.ma.masked_where(tiled_position_mask, position_grid)
        # dense grid
        return position_grid


    def hrir(self, subject_id, side, fundamental_angles=None, orthogonal_angles=None, distance=None):
        sofa_path = self._sofa_path(subject_id)
        hrir_file = ncdf.Dataset(sofa_path)
        try:
            hrirs = np.ma.getdata(hrir_file.variables['Data.IR'][:, 0 if side.endswith('left') else 1, :])
            samplerate = hrir_file.variables['Data.SamplingRate'][:].item()
        except Exception as exc:
            raise ValueError(f'Error reading file "{sofa_path}"') from exc
        finally:
            hrir_file.close()
        selected_fundamental_angles, _, _, selection_mask, selected_file_indices, selection_mask_indices = self._map_sofa_position_order_to_matrix(subject_id, fundamental_angles, orthogonal_angles, distance)
        hrir_matrix = np.empty(selection_mask.shape + (hrirs.shape[1],))
        hrir_matrix[selection_mask_indices] = hrirs[selected_file_indices]
        tiled_selection_mask = np.tile(selection_mask[:, :, :, np.newaxis], (1, 1, 1, hrir_matrix.shape[-1]))
        hrir = np.ma.masked_where(tiled_selection_mask, hrir_matrix, copy=False)
        if side.startswith('mirrored'):
            return self._mirror_hrirs(hrir, selected_fundamental_angles)
        return hrir


class SofaSphericalDataReader(SofaDataReader):

    @property
    def fundamental_angle_name(self):
        return 'azimuth [°]'


    @property
    def orthogonal_angle_name(self):
        return 'elevation [°]'


    def hrir_positions(self, subject_id, fundamental_angles=None, orthogonal_angles=None, coordinate_system='spherical'):
        return super().hrir_positions(subject_id, coordinate_system, fundamental_angles, orthogonal_angles)


    @staticmethod
    def _convert_positions(coordinate_system, positions):
        if coordinate_system == 'cartesian':
            positions = np.stack(cartesian2spherical(*positions.T), axis=1)
        return positions


    @staticmethod
    def _coordinate_transform(coordinate_system, selected_fundamental_angles, selected_orthogonal_angles, selected_radii):
        if coordinate_system == 'spherical':
            return selected_fundamental_angles, selected_orthogonal_angles, selected_radii
        if coordinate_system == 'interaural':
            return spherical2interaural(selected_fundamental_angles, selected_orthogonal_angles, selected_radii)
        if coordinate_system == 'cartesian':
            return spherical2cartesian(selected_fundamental_angles, selected_orthogonal_angles, selected_radii)
        raise ValueError(f'Unknown coordinate system "{coordinate_system}"')


    @staticmethod
    def _mirror_hrirs(hrirs, selected_fundamental_angles):
        # flip azimuths (in rows)
        if np.isclose(selected_fundamental_angles[0], -180):
            return np.ma.row_stack((hrirs[0:1], np.flipud(hrirs[1:])))
        else:
            return np.flipud(hrirs)


    @staticmethod
    def _verify_angle_symmetry(fundamental_angles, _):
        # mirror azimuths/rows
        start_idx = 1 if np.isclose(fundamental_angles[0], -180) else 0
        if not np.allclose(fundamental_angles[start_idx:], -np.flip(fundamental_angles[start_idx:])):
            raise ValueError('Only datasets with symmetric azimuths can mix mirrored and non-mirrored sides.')


class SofaInterauralDataReader(SofaDataReader):

    @property
    def fundamental_angle_name(self):
        return 'vertical [°]'


    @property
    def orthogonal_angle_name(self):
        return 'lateral [°]'


    def hrir_positions(self, subject_id, fundamental_angles=None, orthogonal_angles=None, coordinate_system='interaural'):
        return super().hrir_positions(subject_id, coordinate_system, fundamental_angles, orthogonal_angles)


    @staticmethod
    def _convert_positions(coordinate_system, positions):
        if coordinate_system == 'cartesian':
            positions = np.stack(cartesian2interaural(*positions.T), axis=1)
        else:
            positions = np.stack(spherical2interaural(*positions.T), axis=1)
        positions[:, [0, 1]] = positions[:, [1, 0]]
        return positions


    @staticmethod
    def _coordinate_transform(coordinate_system, selected_fundamental_angles, selected_orthogonal_angles, selected_radii):
        if coordinate_system == 'interaural':
            coordinates = selected_fundamental_angles, selected_orthogonal_angles, selected_radii
        elif coordinate_system == 'spherical':
            coordinates = interaural2spherical(selected_orthogonal_angles, selected_fundamental_angles, selected_radii)
            coordinates[0], coordinates[1] = coordinates[1], coordinates[0]
        elif coordinate_system == 'cartesian':
            coordinates = interaural2cartesian(selected_fundamental_angles, selected_fundamental_angles, selected_radii)
            coordinates[0], coordinates[1] = coordinates[1], coordinates[0]
        else:
            raise ValueError(f'Unknown coordinate system "{coordinate_system}"')
        return coordinates


    @staticmethod
    def _mirror_hrirs(hrirs, _):
        # flip lateral angles (in columns)
        return np.fliplr(hrirs)


    @staticmethod
    def _verify_angle_symmetry(_, orthogonal_angles):
        # mirror laterals/columns
        if not np.allclose(orthogonal_angles, -np.flip(orthogonal_angles)):
            raise ValueError('Only datasets with symmetric lateral angles can mix mirrored and non-mirrored sides.')


class AnthropometryDataReader(DataReader):

    def anthropometric_data(self, subject_id, side, select=None):
        subject_idx = np.squeeze(np.argwhere(np.squeeze(self.query._anthropometric_ids) == subject_id))
        if subject_idx.size == 0:
            raise ValueError(f'Subject id "{subject_id}" has no anthropometric measurements')
        selected_data = self.query._anthropometry_values(side, select)[subject_idx]
        if np.all(np.isnan(selected_data), axis=-1):
            raise ValueError(f'Subject id "{subject_id}" has no data available for selection "{", ".join(select) if select is not None else "None"}"')
        return selected_data
    

class ImageDataReader(DataReader):

    @abstractmethod
    def _image_path(self, subject_id, side, rear=False):
        pass


    def image(self, subject_id, side, rear=False):
        img = Image.open(self._image_path(subject_id, side, rear))
        if side.startswith('mirrored-'):
            return img.transpose(Image.FLIP_LEFT_RIGHT)
        return img


class CipicDataReader(SofaInterauralDataReader, AnthropometryDataReader, ImageDataReader):

    def __init__(self,
        sofa_directory_path: Union[str, Path] = '',
        image_directory_path: Union[str, Path] = '',
        anthropometry_matfile_path: Union[str, Path] = '',
        download: bool = False,
        verify: bool = True,
    ):
        query = CipicDataQuery(sofa_directory_path, image_directory_path, anthropometry_matfile_path, download, verify)
        super().__init__(query)


    def _sofa_path(self, subject_id):
        return str(self.query.sofa_directory_path / f'subject_{subject_id:03d}.sofa')


    def _image_path(self, subject_id, side, rear=False):
        candidate_paths = [self.query.image_directory_path /  f'Subject_{subject_id:03d}' / f'{subject_id:03d}{suffix}' for suffix in self.query._image_suffix(side, rear)]

        for image_path in candidate_paths:
            if image_path.exists():
                return image_path
        raise ValueError(f'No {side} {"rear" if rear else "side"} image available for subject {subject_id}')


class AriDataReader(SofaSphericalDataReader, AnthropometryDataReader):

    def __init__(self,
        sofa_directory_path: Union[str, Path] = '',
        anthropometry_matfile_path: Union[str, Path] = '',
        download: bool = False,
        verify: bool = True,
    ):
        query = AriDataQuery(sofa_directory_path, anthropometry_matfile_path, download, verify)
        super().__init__(query)


    def _sofa_path(self, subject_id):
        try:
            return str(next(self.query.sofa_directory_path.glob(f'hrtf [bc]_nh{subject_id}.sofa')))
        except :
            raise ValueError(f'No subject with id "{subject_id}" exists in the collection')


class ListenDataReader(SofaSphericalDataReader, AnthropometryDataReader):

    def __init__(self,
        sofa_directory_path: Union[str, Path] = '',
        anthropometry_directory_path: Union[str, Path] = '',
        hrir_variant: str = 'compensated',
        download: bool = False,
        verify: bool = True,
    ):
        query = ListenDataQuery(sofa_directory_path, anthropometry_directory_path, hrir_variant, download, verify)
        super().__init__(query)


    def _sofa_path(self, subject_id):
        return str(self.query.sofa_directory_path / self.query._checksum_key / '44100' / f'IRC_{subject_id:04d}_{self.query._hrir_variant_char}_44100.sofa')


class CrossModDataReader(SofaSphericalDataReader):

    def __init__(self,
        sofa_directory_path: Union[str, Path] = '',
        hrir_variant: str = 'compensated',
        download: bool = False,
        verify: bool = True,
    ):
        query = CrossModDataQuery(sofa_directory_path, hrir_variant, download, verify)
        super().__init__(query)


    def _sofa_path(self, subject_id):
        return str(self.query.sofa_directory_path / self.query._checksum_key / '44100' / f'IRC_{subject_id:04d}_{self.query._hrir_variant_char}_44100.sofa')


class BiLiDataReader(SofaSphericalDataReader):

    def __init__(self,
        sofa_directory_path: Union[str, Path] = '',
        hrir_variant: str = 'compensated',
        hrir_samplerate: Optional[float] = None,
        download: bool = False,
        verify: bool = True,
    ):
        query = BiLiDataQuery(sofa_directory_path, hrir_samplerate, hrir_variant, download, verify)
        super().__init__(query)


    def _sofa_path(self, subject_id):
        return str(self.query.sofa_directory_path / self.query._hrir_variant / str(self.query._samplerate) / f'IRC_{subject_id:04d}_{self.query._hrir_variant_char}_HRIR_{self.query._samplerate}.sofa')


class ItaDataReader(SofaSphericalDataReader, AnthropometryDataReader):

    def __init__(self,
        sofa_directory_path: Union[str, Path] = '',
        anthropometry_csvfile_path: Union[str, Path] = '',
        download: bool = False,
        verify: bool = True,
    ):
        query = ItaDataQuery(sofa_directory_path, anthropometry_csvfile_path, download, verify)
        super().__init__(query)


    def _sofa_path(self, subject_id):
        return str(self.query.sofa_directory_path / f'MRT{subject_id:02d}.sofa')


class HutubsDataReader(SofaSphericalDataReader, AnthropometryDataReader):

    def __init__(self,
        sofa_directory_path: Union[str, Path] = '',
        anthropometry_csvfile_path: Union[str, Path] = '',
        hrir_method: Optional[str] = None,
        download: bool = False,
        verify: bool = True,
    ):
        query = HutubsDataQuery(sofa_directory_path, anthropometry_csvfile_path, hrir_method, download, verify)
        super().__init__(query)


    def _sofa_path(self, subject_id):
        return str(self.query.sofa_directory_path / f'pp{subject_id:d}_HRIRs_{self.query._method_str}.sofa')


class RiecDataReader(SofaSphericalDataReader):

    def __init__(self,
        sofa_directory_path: Union[str, Path] = '',
        download: bool = False,
        verify: bool = True,
    ):
        query = RiecDataQuery(sofa_directory_path, download, verify)
        super().__init__(query)


    def _sofa_path(self, subject_id):
        return str(self.query.sofa_directory_path / f'RIEC_hrir_subject_{subject_id:03d}.sofa')


class ChedarDataReader(SofaSphericalDataReader, AnthropometryDataReader):

    _angle_quantisation: int = 5


    def __init__(self,
        sofa_directory_path: Union[str, Path] = '',
        anthropometry_matfile_path: Union[str, Path] = '',
        distance: Optional[Union[float, str]] = None,
        download: bool = False,
        verify: bool = True,
    ):
        query = ChedarDataQuery(sofa_directory_path, anthropometry_matfile_path, distance, download, verify)
        super().__init__(query)


    def _sofa_path(self, subject_id):
        return str(self.query.sofa_directory_path / self.query._radius / f'chedar_{subject_id:04d}_UV{self.query._radius}.sofa')


class WidespreadDataReader(SofaSphericalDataReader):

    def __init__(self,
        sofa_directory_path: Union[str, Path] = '',
        distance: Optional[Union[float, str]] = None,
        grid: str = 'UV',
        download: bool = False,
        verify: bool = True,
    ):
        query = WidespreadDataQuery(sofa_directory_path, distance, grid, download, verify)
        super().__init__(query)
        if grid == 'UV':
            self._angle_quantisation = 5
        else:
            self._angle_quantisation = 0

    def _sofa_path(self, subject_id):
        return str(self.query.sofa_directory_path / self.query._grid / self.query._radius / f'{self.query._grid}{self.query._radius}_{subject_id:05d}.sofa')


class Sadie2DataReader(SofaSphericalDataReader, ImageDataReader):

    def __init__(self,
        sofa_directory_path: Union[str, Path] = '',
        image_directory_path: Union[str, Path] = '',
        hrir_samplerate: Optional[float] = None,
        download: bool = False,
        verify: bool = True,
    ):
        query = Sadie2DataQuery(sofa_directory_path, image_directory_path, hrir_samplerate, download, verify)
        super().__init__(query)


    def _sofa_path(self, subject_id):
        if subject_id < 3:
            sadie2_id = f'D{subject_id}'
        else:
            sadie2_id = f'H{subject_id}'
        return str(self.query.sofa_directory_path / sadie2_id / f'{sadie2_id}_HRIR_SOFA' / f'{sadie2_id}_{self.query._samplerate_str}_FIR_SOFA.sofa')


    def _image_path(self, subject_id, side, rear=False):
        if rear:
            raise ValueError('No rear pictures available in the SADIE II dataset')
        side_str = self.query._image_side_str(side)
        if subject_id < 3:
            sadie2_id = f'D{subject_id}'
            side_str = ' ' + side_str
        else:
            sadie2_id = f'H{subject_id}'
            side_str = '_' + side_str
        return str(self.query.image_directory_path / sadie2_id / f'{sadie2_id}_Scans' / f'{sadie2_id}{side_str}.png')


class Princeton3D3ADataReader(SofaSphericalDataReader, AnthropometryDataReader):

    def __init__(self,
        sofa_directory_path: Union[str, Path] = '',
        anthropometry_directory_path: Union[str, Path] = '',
        hrir_method: Optional[str] = None,
        hrir_variant: str = 'compensated',
        download: bool = False,
        verify: bool = True,
    ):
        query = Princeton3D3ADataQuery(sofa_directory_path, anthropometry_directory_path, hrir_method, hrir_variant, download, verify)
        super().__init__(query)


    def _sofa_path(self, subject_id):
        return str(self.query.sofa_directory_path / self.query._method_str / f'Subject{subject_id}' / f'Subject{subject_id}_{self.query._hrir_variant_str}.sofa')



class ScutDataReader(SofaSphericalDataReader, AnthropometryDataReader):

    def __init__(self,
        sofa_directory_path: Union[str, Path] = '',
        anthropometry_csvfile_path: Union[str, Path] = '',
        download: bool = False,
        verify: bool = True,
    ):
        query = ScutDataQuery(sofa_directory_path, anthropometry_csvfile_path, download, verify)
        super().__init__(query)


    def _sofa_path(self, subject_id):
        return str(self.query.sofa_directory_path / f'SCUT_NF_subject{subject_id:04d}_measured.sofa')


class SonicomDataReader(SofaSphericalDataReader):

    def __init__(self,
        sofa_directory_path: Union[str, Path] = '',
        hrir_variant: str = 'compensated',
        hrir_samplerate: Optional[float] = None,
        download: bool = False,
        verify: bool = True,
    ):
        query = SonicomDataQuery(sofa_directory_path, hrir_samplerate, hrir_variant, download, verify)
        super().__init__(query)


    def _sofa_path(self, subject_id):
        if isinstance(subject_id, int):
            base_dir = f'P{subject_id:04d}'
        else:
            base_dir = subject_id
        return str(self.query.sofa_directory_path / base_dir / 'HRTF' / 'HRTF' / self.query._samplerate_str / f'{base_dir}_{self.query._hrir_variant_str}_{self.query._samplerate_str}.sofa')


class MitKemarDataReader(SofaSphericalDataReader):

    def __init__(self,
        sofa_directory_path: Union[str, Path] = '',
        download: bool = False,
        verify: bool = True,
    ):
        query = MitKemarDataQuery(sofa_directory_path, download, verify)
        super().__init__(query)


    def _sofa_path(self, subject_id):
        return str(self.query.sofa_directory_path / f'mit_kemar_{subject_id}_pinna.sofa')
